@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2008 Jeffrey Stedfast
+ *  Copyright (C) 2000-2010 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -59,17 +59,17 @@
 struct _GMimeHeader {
 	GMimeHeader *next;
 	GMimeHeader *prev;
-	gint64 offset;
+	/*gint64 offset;*/
 	char *name;
 	char *value;
 };
 
 struct _GMimeHeaderList {
+	GMimeStream *stream;
 	GHashTable *writers;
 	GHashTable *hash;
 	guint32 version;
 	List list;
-	char *raw;
 };
 
 
@@ -92,10 +92,10 @@ g_mime_header_new (const char *name, const char *value, gint64 offset)
 {
 	GMimeHeader *header;
 	
-	header = g_new (GMimeHeader, 1);
+	header = g_slice_new (GMimeHeader);
 	header->name = g_strdup (name);
 	header->value = g_strdup (value);
-	header->offset = offset;
+	/*header->offset = offset;*/
 	header->next = NULL;
 	header->prev = NULL;
 	
@@ -114,7 +114,8 @@ g_mime_header_free (GMimeHeader *header)
 {
 	g_free (header->name);
 	g_free (header->value);
-	g_free (header);
+	
+	g_slice_free (GMimeHeader, header);
 }
 
 
@@ -133,7 +134,7 @@ g_mime_header_iter_new (void)
 {
 	GMimeHeaderIter *iter;
 	
-	iter = g_new (GMimeHeaderIter, 1);
+	iter = g_slice_new (GMimeHeaderIter);
 	iter->hdrlist = NULL;
 	iter->cursor = NULL;
 	iter->version = 0;
@@ -194,7 +195,7 @@ g_mime_header_iter_free (GMimeHeaderIter *iter)
 {
 	g_return_if_fail (iter != NULL);
 	
-	g_free (iter);
+	g_slice_free (GMimeHeaderIter, iter);
 }
 
 
@@ -235,10 +236,13 @@ g_mime_header_iter_is_valid (GMimeHeaderIter *iter)
 {
 	g_return_val_if_fail (iter != NULL, FALSE);
 	
-	if (!iter->hdrlist || !iter->cursor || !iter->cursor->next)
+	if (!iter->hdrlist || iter->version != iter->hdrlist->version)
 		return FALSE;
 	
-	return iter->version == iter->hdrlist->version;
+	if (!iter->cursor || !iter->cursor->next)
+		return FALSE;
+	
+	return TRUE;
 }
 
 
@@ -362,6 +366,7 @@ g_mime_header_iter_prev (GMimeHeaderIter *iter)
 }
 
 
+#if 0
 /**
  * g_mime_header_iter_get_offset:
  * @iter: a #GMimeHeaderIter
@@ -380,6 +385,7 @@ g_mime_header_iter_get_offset (GMimeHeaderIter *iter)
 	
 	return iter->cursor->offset;
 }
+#endif
 
 
 /**
@@ -423,7 +429,7 @@ g_mime_header_iter_set_value (GMimeHeaderIter *iter, const char *value)
 	g_free (iter->cursor->value);
 	iter->cursor->value = g_strdup (value);
 	
-	g_mime_header_list_set_raw (iter->hdrlist, NULL);
+	g_mime_header_list_set_stream (iter->hdrlist, NULL);
 	
 	return TRUE;
 }
@@ -521,15 +527,15 @@ g_mime_header_list_new (void)
 {
 	GMimeHeaderList *headers;
 	
-	headers = g_new (GMimeHeaderList, 1);
+	headers = g_slice_new (GMimeHeaderList);
 	headers->writers = g_hash_table_new_full (g_mime_strcase_hash,
 						  g_mime_strcase_equal,
 						  g_free, NULL);
 	headers->hash = g_hash_table_new (g_mime_strcase_hash,
 					  g_mime_strcase_equal);
 	list_init (&headers->list);
+	headers->stream = NULL;
 	headers->version = 0;
-	headers->raw = NULL;
 	
 	return headers;
 }
@@ -558,8 +564,11 @@ g_mime_header_list_destroy (GMimeHeaderList *headers)
 	
 	g_hash_table_destroy (headers->writers);
 	g_hash_table_destroy (headers->hash);
-	g_free (headers->raw);
-	g_free (headers);
+	
+	if (headers->stream)
+		g_object_unref (headers->stream);
+	
+	g_slice_free (GMimeHeaderList, headers);
 }
 
 
@@ -585,8 +594,10 @@ g_mime_header_list_prepend (GMimeHeaderList *headers, const char *name, const ch
 	list_append (&headers->list, (ListNode *) header);
 	g_hash_table_replace (headers->hash, header->name, header);
 	
-	g_free (headers->raw);
-	headers->raw = NULL;
+	if (headers->stream) {
+		g_object_unref (headers->stream);
+		headers->stream = NULL;
+	}
 }
 
 
@@ -614,8 +625,10 @@ g_mime_header_list_append (GMimeHeaderList *headers, const char *name, const cha
 	if (!g_hash_table_lookup (headers->hash, name))
 		g_hash_table_insert (headers->hash, header->name, header);
 	
-	g_free (headers->raw);
-	headers->raw = NULL;
+	if (headers->stream) {
+		g_object_unref (headers->stream);
+		headers->stream = NULL;
+	}
 }
 
 
@@ -649,16 +662,19 @@ g_mime_header_list_get (const GMimeHeaderList *headers, const char *name)
  * @name: header name
  * @value: header value
  *
- * Set the value of the first header with the name specified. If
- * @value is %NULL and the header, @name, had not been previously set,
- * a space will be set aside for it (useful for setting the order of
- * headers before values can be obtained for them) otherwise the
- * header will be unset.
+ * Set the value of the specified header. If @value is %NULL and the
+ * header, @name, had not been previously set, a space will be set
+ * aside for it (useful for setting the order of headers before values
+ * can be obtained for them) otherwise the header will be unset.
+ *
+ * Note: If there are multiple headers with the specified field name,
+ * the first instance of the header will be replaced and further
+ * instances will be removed.
  **/
 void
 g_mime_header_list_set (GMimeHeaderList *headers, const char *name, const char *value)
 {
-	GMimeHeader *header;
+	GMimeHeader *header, *next;
 	
 	g_return_if_fail (headers != NULL);
 	g_return_if_fail (name != NULL);
@@ -666,14 +682,30 @@ g_mime_header_list_set (GMimeHeaderList *headers, const char *name, const char *
 	if ((header = g_hash_table_lookup (headers->hash, name))) {
 		g_free (header->value);
 		header->value = g_strdup (value);
+		
+		header = header->next;
+		while (header->next) {
+			next = header->next;
+			
+			if (!g_ascii_strcasecmp (header->name, name)) {
+				/* remove/free the header */
+				list_unlink ((ListNode *) header);
+				g_mime_header_free (header);
+				headers->version++;
+			}
+			
+			header = next;
+		}
 	} else {
 		header = g_mime_header_new (name, value, -1);
 		list_append (&headers->list, (ListNode *) header);
 		g_hash_table_insert (headers->hash, header->name, header);
 	}
 	
-	g_free (headers->raw);
-	headers->raw = NULL;
+	if (headers->stream) {
+		g_object_unref (headers->stream);
+		headers->stream = NULL;
+	}
 }
 
 
@@ -692,8 +724,8 @@ g_mime_header_list_remove (GMimeHeaderList *headers, const char *name)
 {
 	GMimeHeader *header, *node;
 	
-	g_return_if_fail (headers != NULL);
-	g_return_if_fail (name != NULL);
+	g_return_val_if_fail (headers != NULL, FALSE);
+	g_return_val_if_fail (name != NULL, FALSE);
 	
 	if (!(header = g_hash_table_lookup (headers->hash, name)))
 		return FALSE;
@@ -717,8 +749,12 @@ g_mime_header_list_remove (GMimeHeaderList *headers, const char *name)
 	list_unlink ((ListNode *) header);
 	g_mime_header_free (header);
 	
-	g_free (headers->raw);
-	headers->raw = NULL;
+	if (headers->stream) {
+		g_object_unref (headers->stream);
+		headers->stream = NULL;
+	}
+	
+	return TRUE;
 }
 
 
@@ -751,6 +787,31 @@ g_mime_header_list_get_iter (GMimeHeaderList *headers, GMimeHeaderIter *iter)
 }
 
 
+/**
+ * g_mime_header_list_foreach:
+ * @headers: A #GMimeHeaderList
+ * @func: function to be called for each header.
+ * @user_data: User data to be passed to the func.
+ *
+ * Calls @func for each header name/value pair.
+ */
+void
+g_mime_header_list_foreach (const GMimeHeaderList *headers, GMimeHeaderForeachFunc func, gpointer user_data)
+{
+	const GMimeHeader *header;
+	
+	g_return_if_fail (headers != NULL);
+	g_return_if_fail (func != NULL);
+	
+	header = (const GMimeHeader *) headers->list.head;
+	
+	while (header->next) {
+		func (header->name, header->value, user_data);
+		header = header->next;
+	}
+}
+
+
 static ssize_t
 default_writer (GMimeStream *stream, const char *name, const char *value)
 {
@@ -785,8 +846,10 @@ g_mime_header_list_write_to_stream (const GMimeHeaderList *headers, GMimeStream
 	g_return_val_if_fail (headers != NULL, -1);
 	g_return_val_if_fail (stream != NULL, -1);
 	
-	if (headers->raw)
-		return g_mime_stream_write_string (stream, headers->raw);
+	if (headers->stream) {
+		g_mime_stream_reset (headers->stream);
+		return g_mime_stream_write_to_stream (headers->stream, stream);
+	}
 	
 	header = (GMimeHeader *) headers->list.head;
 	writers = headers->writers;
@@ -796,7 +859,7 @@ g_mime_header_list_write_to_stream (const GMimeHeaderList *headers, GMimeStream
 			if (!(writer = g_hash_table_lookup (writers, header->name)))
 				writer = default_writer;
 			
-			if ((nwritten = (* writer) (stream, header->name, header->value)) == -1)
+			if ((nwritten = writer (stream, header->name, header->value)) == -1)
 				return -1;
 			
 			total += nwritten;
@@ -827,13 +890,17 @@ g_mime_header_list_to_string (const GMimeHeaderList *headers)
 	
 	g_return_val_if_fail (headers != NULL, NULL);
 	
-	if (headers->raw)
-		return g_strdup (headers->raw);
-	
 	array = g_byte_array_new ();
 	stream = g_mime_stream_mem_new ();
 	g_mime_stream_mem_set_byte_array (GMIME_STREAM_MEM (stream), array);
-	g_mime_header_list_write_to_stream (headers, stream);
+	
+	if (headers->stream) {
+		g_mime_stream_reset (headers->stream);
+		g_mime_stream_write_to_stream (headers->stream, stream);
+	} else {
+		g_mime_header_list_write_to_stream (headers, stream);
+	}
+	
 	g_object_unref (stream);
 	
 	g_byte_array_append (array, (unsigned char *) "", 1);
@@ -874,20 +941,71 @@ g_mime_header_list_register_writer (GMimeHeaderList *headers, const char *name,
 
 
 
+/**
+ * g_mime_header_list_set_stream:
+ * @headers: a #GMimeHeaderList
+ * @stream: a #GMimeStream
+ *
+ * Set the raw header stream.
+ *
+ * Since: 2.4.7
+ **/
+void
+g_mime_header_list_set_stream (GMimeHeaderList *headers, GMimeStream *stream)
+{
+	g_return_if_fail (stream == NULL || GMIME_IS_STREAM (stream));
+	g_return_if_fail (headers != NULL);
+	
+	if (stream)
+		g_object_ref (stream);
+	
+	if (headers->stream)
+		g_object_unref (headers->stream);
+	
+	headers->stream = stream;
+}
+
+
+/**
+ * g_mime_header_list_get_stream:
+ * @headers: a #GMimeHeaderList
+ *
+ * Gets the raw stream representing @headers.
+ *
+ * Returns: a #GMimeStream if set or %NULL otherwise.
+ *
+ * Since: 2.4.7
+ **/
+GMimeStream *
+g_mime_header_list_get_stream (GMimeHeaderList *headers)
+{
+	g_return_val_if_fail (headers != NULL, NULL);
+	
+	return headers->stream;
+}
+
+
 /**
  * g_mime_header_list_set_raw:
  * @headers: a #GMimeHeaderList
  * @raw: raw mime part header
  *
  * Set the raw header.
+ *
+ * Deprecated:2.4.7: Use g_mime_header_list_set_stream() instead.
  **/
 void
 g_mime_header_list_set_raw (GMimeHeaderList *headers, const char *raw)
 {
 	g_return_if_fail (headers != NULL);
 	
-	g_free (headers->raw);
-	headers->raw = raw ? g_strdup (raw) : NULL;
+	if (headers->stream)
+		g_object_unref (headers->stream);
+	
+	if (raw)
+		headers->stream = g_mime_stream_mem_new_with_buffer (raw, strlen (raw));
+	else
+		headers->stream = NULL;
 }
 
 
@@ -898,11 +1016,14 @@ g_mime_header_list_set_raw (GMimeHeaderList *headers, const char *raw)
  * Gets whether or not a raw header has been set on @headers.
  *
  * Returns: %TRUE if a raw header is set or %FALSE otherwise.
+ *
+ * Deprecated:2.4.7: Use g_mime_header_list_get_stream() and check
+ * against NULL instead.
  **/
 gboolean
 g_mime_header_list_has_raw (const GMimeHeaderList *headers)
 {
 	g_return_val_if_fail (headers != NULL, FALSE);
 	
-	return headers->raw ? TRUE : FALSE;
+	return headers->stream ? TRUE : FALSE;
 }
